# flake8: noqa

from .account import GetAccountArgs
from .block import (
    AlchemyBlockReceipt,
    AlchemyParams,
    BlockNumberArg,
    GetBlockByHashArgs,
    GetBlockByNumberArgs,
)
from .eth_call import CallWithBlockArgs, EthCallArgs, EthCallParams
from .fee_history import FeeHistoryArgs
from .logs import LogsArgs, LogsParams
from .storage import GetCodeArgs, GetStorageArgs
from .tracer import TraceArgs, TracerConfig
from .transaction import (
    GetTransactionByBlockHash,
    GetTransactionByBlockNumber,
    RawTransaction,
    TransactionRequest,
)



from inspect import isclass
from types import GenericAlias
from typing import NamedTuple, Optional, get_args, get_origin

from eth_abi import encode
from eth_rpc.types import primitives
from eth_typing import HexStr
from eth_utils import to_bytes, to_hex
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from ..types import Name, map_type_to_str
from .types import is_annotation


def encode_to_string(word):
    return encode(("string",), (word,))


def encode_to_bytes(word: str, type: primitives.BYTES_TYPES = primitives.bytes32):
    return encode((type.__name__,), (word.encode("utf-8"),))


def hex_to_int(v: None | int | str) -> Optional[int]:
    if v is None:
        return None
    return int(v, 16) if isinstance(v, str) else v


def to_32byte_hex(val):
    return to_hex(to_bytes(val).rjust(32, b"\0"))


def to_hex_str(number: int | str) -> HexStr:
    if isinstance(number, int):
        return HexStr(hex(number))
    return HexStr(number)


def convert_base_model(
    base: type[BaseModel], with_name: bool = False, as_tuple: bool = False
):
    lst = []
    for key, field_info in base.model_fields.items():
        lst.append(_convert_field_info(key, field_info, with_name=with_name))
    if as_tuple:
        return ",".join(lst)
    return lst


def _convert_field_info(alias: str, field: FieldInfo, with_name: bool = False):
    name = "" if not with_name else alias
    field_type = field.annotation

    for metadata in field.metadata:
        if isinstance(metadata, Name) and with_name:
            name = metadata.value

    type_origin = get_origin(field_type)
    if type_origin == list:
        list_type = get_args(field_type)[0]
        converted_list_type = convert(list_type)
        if isinstance(converted_list_type, list):
            converted_list_type = f"({','.join(converted_list_type)})"
        return f"{converted_list_type}[] {name}".strip()
    elif type_origin == tuple:
        tuple_args = f"({','.join([convert(t) for t in get_args(field_type)])}) {name}"
        return tuple_args
    return f"{map_type_to_str(field_type)} {name}".strip()


def convert(type, with_name: bool = False):
    name = ""
    # unwrap annotations
    if is_annotation(type):
        type, *annotations = get_args(type)
        if with_name:
            for annotation in annotations:
                if isinstance(annotation, Name):
                    name = annotation.value

    # if it's a list or a tuple
    if isinstance(type, GenericAlias):
        if get_origin(type) in [list, "list"]:
            list_type = get_args(type)[0]
            converted_list_type = convert(list_type)
            # if the list type is a tuple:
            if isinstance(converted_list_type, list):
                converted_list_type = f"({','.join(converted_list_type)})"
            return f"{converted_list_type}[] {name}".strip()
        else:
            tuple_args = [convert(t, with_name=with_name) for t in get_args(type)]
            return tuple_args
    elif getattr(type, "__orig_bases__", [None])[0] == NamedTuple:
        return f"({','.join([convert(t) for t in type.__annotations__.values()])}) {name}".strip()
    elif isclass(type) and issubclass(type, BaseModel):
        return f"({convert_base_model(type, with_name=with_name, as_tuple=True)}) {name}".strip()
    return f"{map_type_to_str(type)} {name}".strip()



from typing import Any

from eth_rpc.models import (
    AccessListResponse,
    Account,
    Block,
    FeeHistory,
    Log,
    PendingTransaction,
)
from eth_rpc.transaction import AlchemyReceiptsResponse
from eth_rpc.types import (
    AlchemyBlockReceipt,
    BlockNumberArg,
    CallWithBlockArgs,
    EthCallArgs,
    FeeHistoryArgs,
    GetAccountArgs,
    GetBlockByHashArgs,
    GetBlockByNumberArgs,
    GetCodeArgs,
    GetStorageArgs,
    GetTransactionByBlockHash,
    GetTransactionByBlockNumber,
    HexInteger,
    LogsArgs,
    NoArgs,
    RawTransaction,
    TraceArgs,
    TransactionRequest,
)
from eth_typing import HexStr

from ..models import Transaction as TransactionModel
from ..models import TransactionReceipt as TransactionReceiptModel
from .base import BaseRPC
from .method import RPCMethod


class RPC(BaseRPC):
    chain_id: RPCMethod = RPCMethod[NoArgs, HexInteger](name="eth_chainId")
    max_priority_fee_per_gas: RPCMethod = RPCMethod[NoArgs, HexInteger](
        name="eth_maxPriorityFeePerGas"
    )
    fee_history: RPCMethod = RPCMethod[FeeHistoryArgs, FeeHistory](
        name="eth_feeHistory"
    )
    eth_call: RPCMethod = RPCMethod[EthCallArgs, HexStr](name="eth_call")
    get_block_by_hash: RPCMethod = RPCMethod[GetBlockByHashArgs, Block](
        name="eth_getBlockByHash"
    )
    get_block_by_number: RPCMethod = RPCMethod[GetBlockByNumberArgs, Block](
        name="eth_getBlockByNumber"
    )
    get_block_tx_count_by_number: RPCMethod = RPCMethod[BlockNumberArg, HexInteger](
        name="eth_getBlockTransactionCountByNumber",
    )
    get_tx_by_hash: RPCMethod = RPCMethod[TransactionRequest, TransactionModel | None](
        name="eth_getTransactionByHash"
    )
    get_pending_tx_by_hash: RPCMethod = RPCMethod[
        TransactionRequest, PendingTransaction | None
    ](name="eth_getTransactionByHash")
    estimate_gas: RPCMethod = RPCMethod[CallWithBlockArgs, HexInteger](
        name="eth_estimateGas"
    )
    block_number: RPCMethod = RPCMethod[NoArgs, HexInteger](name="eth_blockNumber")
    create_access_list: RPCMethod = RPCMethod[CallWithBlockArgs, AccessListResponse](
        name="eth_createAccessList"
    )
    get_storage_at: RPCMethod = RPCMethod[GetStorageArgs, HexStr](
        name="eth_getStorageAt"
    )
    get_code: RPCMethod = RPCMethod[GetCodeArgs, HexStr](name="eth_getCode")
    get_tx_receipt: RPCMethod = RPCMethod[
        TransactionRequest, TransactionReceiptModel | None
    ](name="eth_getTransactionReceipt")
    get_block_receipts: RPCMethod = RPCMethod[
        list[HexInteger | HexStr], list[TransactionReceiptModel]
    ](name="eth_getBlockReceipts")
    alchemy_get_block_receipts: RPCMethod = RPCMethod[
        AlchemyBlockReceipt, AlchemyReceiptsResponse
    ](name="alchemy_getTransactionReceipts")
    get_tx_by_block_hash: RPCMethod = RPCMethod[
        GetTransactionByBlockHash, TransactionModel
    ](name="eth_getTransactionByBlockHashAndIndex")
    get_tx_by_block_number: RPCMethod = RPCMethod[
        GetTransactionByBlockNumber, TransactionModel
    ](name="eth_getTransactionByBlockNumberAndIndex")
    get_balance: RPCMethod = RPCMethod[GetAccountArgs, HexInteger](
        name="eth_getBalance"
    )
    get_account: RPCMethod = RPCMethod[GetAccountArgs, Account](name="eth_getAccount")
    get_tx_count: RPCMethod = RPCMethod[GetAccountArgs, HexInteger](
        name="eth_getTransactionCount"
    )
    get_logs: RPCMethod = RPCMethod[LogsArgs, list[Log]](name="eth_getLogs")

    send_raw_tx: RPCMethod = RPCMethod[RawTransaction, HexStr](
        name="eth_sendRawTransaction"
    )

    # debug methods
    debug_tracecall: RPCMethod = RPCMethod[TraceArgs, Any](name="debug_traceCall")

    def model_post_init(self, __context):
        rpc_methods = []
        for method in dir(self):
            if method[0] == "_":
                continue
            try:
                attr = getattr(self, method)
                if isinstance(attr, RPCMethod):
                    rpc_methods.append((method, attr))
            except Exception:
                pass

        for name, method in rpc_methods:
            new_method = method.copy()
            setattr(self, name, new_method.set_rpc(self).set_network(self.network))

        return super().model_post_init(__context)



import os
from contextvars import ContextVar
from typing import TYPE_CHECKING

from pydantic import BaseModel, ConfigDict

from .exceptions import UnsupportedChainIDException
from .networks import Ethereum, Networks, get_network_by_chain_id
from .rpc.base import BaseRPC
from .types import Network

if TYPE_CHECKING:
    from .rpc.core import RPC


class Transports(BaseModel):
    default: Network = Ethereum
    networks: dict[int, Network] = {}
    rpcs: dict[int, BaseRPC] = {}
    retries: int = 0
    id: int = 0

    def get_network(self, id: int | None = None) -> Network | None:
        if not id:
            return self.default
        return self.networks.get(id)

    def force_get_network(self, id: int | None = None) -> Network:
        network = self.get_network(id=id)
        if not network:
            raise ValueError("transport not found")
        return network

    model_config = ConfigDict(arbitrary_types_allowed=True)


_selected_transports: ContextVar["Transports"] = ContextVar(
    "_selected_transports",
    default=Transports(),
)


def _force_get_default_network() -> Network:
    transport = _selected_transports.get()
    return transport.default


def _force_get_global_rpc(network: Network | None = None) -> "RPC":
    from .rpc.core import RPC

    transport = _selected_transports.get()

    if not network:
        network = transport.default

    if network.chain_id in transport.rpcs:
        return transport.rpcs[network.chain_id]  # type: ignore
    transport.rpcs[network.chain_id] = RPC(network=network)
    return transport.rpcs[network.chain_id]  # type: ignore


def set_transport(networks: list[Network], retries: int = 0, set_default: bool = False):
    transport = _selected_transports.get()
    if set_default:
        set_default_network(networks[0])
    transport.networks.update({network.chain_id: network for network in networks})
    transport.retries = retries


def set_default_network(network: Network):
    transport = _selected_transports.get()
    transport.default = network
    transport.networks[network.chain_id] = network


def get_current_network() -> Network:
    transport = _selected_transports.get()
    return transport.default


def set_rpc_timeout(timeout: float, network: Network | None = None) -> None:
    rpc = _force_get_global_rpc(network)
    rpc.set_timeout(timeout)


def set_alchemy_transport(alchemy_key: str, network: Network):
    set_transport(
        networks=[
            network.set(
                http=f"https://{network.alchemy_str}.g.alchemy.com/v2/{alchemy_key}",
                wss=f"wss://{network.alchemy_str}.g.alchemy.com/v2/{alchemy_key}",
            )
        ],
    )


def set_alchemy_key(alchemy_key: str, network: Network | None = None) -> None:
    """
    Set Alchemy as the rpc url.
    """
    # TODO: it'd be nice to set this for all networks if no network is provided
    if network and not network.alchemy_str:
        raise ValueError("Network not supported by alchemy, set Network.alchemy_str")
    if network:
        set_alchemy_transport(alchemy_key, network)
    else:
        for network in Networks.values():
            set_alchemy_transport(alchemy_key, network)


def configure_rpc_from_env():
    chain_id = int(os.getenv("CHAIN_ID", 1))
    alchemy_key = os.getenv("ALCHEMY_KEY")

    if not alchemy_key:
        raise EnvironmentError(
            "No Alchemy key set. Please set the ALCHEMY_KEY environment variable."
        )

    network = get_network_by_chain_id(chain_id)
    if not network:
        raise UnsupportedChainIDException(f"Unsupported chain ID: {chain_id}")

    set_alchemy_key(alchemy_key, network=network)
    set_default_network(network)



from typing import Any

from eth_typing import HexAddress, HexStr
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel

from ..basic import BlockReference


class LogsParams(BaseModel):
    address: HexAddress | list[HexAddress] | None = None
    from_block: BlockReference
    to_block: BlockReference
    topics: list[list[HexStr] | HexStr | None] | None = None

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )


class LogsArgs(BaseModel):
    params: LogsParams

    def model_dump(self, exclude_none=True, by_alias=True, **kwargs) -> dict[str, Any]:
        return super().model_dump(
            exclude_none=exclude_none, by_alias=by_alias, **kwargs
        )



import pytest
from eth_rpc import Account
from eth_rpc.networks import Arbitrum, Ethereum


@pytest.mark.unit
def test_account_network() -> None:
    # Currently, block is a singleton
    acct = Account[Arbitrum]
    assert acct._network == Arbitrum

    acct2 = Account[Ethereum]
    assert acct._network == Arbitrum
    assert acct2._network == Ethereum



from typing import Generic, Optional, TypeVar

from eth_rpc.types import Network
from eth_rpc.utils import RPCModel
from pydantic import BaseModel, Field

T = TypeVar("T")


class RPCResponse(BaseModel, Generic[T]):
    jsonrpc: float
    id: int
    result: T
    network: Optional[Network] = Field(None)

    def model_post_init(self, __context):
        if isinstance(self.result, RPCModel):
            self.result.set_network(self.network)



import re
from datetime import datetime

import pytz


def convert_datetime_to_iso_8601(dt: datetime) -> str:
    return dt.strftime("%Y-%m-%dT%H:%M:%S")


def load_datetime_string(dt: str | int) -> datetime:
    if isinstance(dt, int):
        return datetime.fromtimestamp(int(dt), tz=pytz.UTC)
    elif re.fullmatch(r"^0x[0-9a-f]+$", dt):
        return datetime.fromtimestamp(int(dt, 16), tz=pytz.UTC)
    elif dt.replace(".", "", 1).isdigit():
        return datetime.fromtimestamp(int(dt), tz=pytz.UTC)
    return datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=pytz.UTC)



from typing import Any, Optional

from eth_typing import HexAddress, HexStr
from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel

from ..basic import BlockReference, HexInteger


class StateOverride(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    balance: Optional[HexInteger] = None
    nonce: Optional[int] = None
    code: Optional[HexStr] = None
    state: Optional[dict[HexStr, HexStr]] = None
    state_diff: Optional[dict[HexStr, HexStr]] = None

    def model_dump(self, exclude_none=True, by_alias=True, **kwargs) -> dict[str, Any]:
        return super().model_dump(
            exclude_none=exclude_none, by_alias=by_alias, **kwargs
        )


class EthCallParams(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    from_: Optional[HexAddress] = Field(default=None, serialization_alias="from")
    to: HexAddress
    gas: Optional[HexInteger] = None
    gas_price: Optional[HexInteger] = None
    value: Optional[HexInteger] = None
    data: Optional[HexStr] = None

    def model_dump(self, exclude_none=True, **kwargs) -> dict[str, Any]:
        return super().model_dump(exclude_none=exclude_none, **kwargs)


class EthCallArgs(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    params: EthCallParams
    block_number: BlockReference
    state_override_set: Optional[dict[HexAddress, StateOverride]] = None

    def model_dump(self, **kwargs) -> dict[str, Any]:
        params = self.params.model_dump(exclude_none=True, by_alias=True)
        block_number = (
            hex(self.block_number)
            if isinstance(self.block_number, int)
            else self.block_number
        )
        if self.state_override_set:
            state_override = {
                key: value.model_dump()
                for key, value in self.state_override_set.items()
            }
            return {
                "params": params,
                "blockNumber": block_number,
                "stateOverride": state_override,
            }
        return {"params": params, "blockNumber": block_number}


class CallWithBlockArgs(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    params: EthCallParams
    block_number: BlockReference | None

    def model_dump(self, exclude_none=True, by_alias=True, **kwargs) -> dict[str, Any]:
        return super().model_dump(
            exclude_none=exclude_none, by_alias=by_alias, **kwargs
        )



from eth_rpc.types import BlockExplorer, Network, Rpcs, RpcUrl
from pydantic.networks import Url

Base = Network(
    chain_id=8453,
    name="Base",
    native_currency="ETH",
    rpc=Rpcs(
        default=RpcUrl(
            http=Url("https://base.llamarpc.com	"),
            wss=Url("wss://base-rpc.publicnode.com"),
        )
    ),
    block_explorer=BlockExplorer(
        name="Basescan",
        url="https://basescan.org",
        api_url="https://api.basescan.org/api",
    ),
    alchemy_str="base-mainnet",
    apprx_block_time=2,
)



from typing import Optional

from eth_rpc.types import HexInteger
from eth_rpc.utils import BloomFilter
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel

from .log import Log


class TransactionReceipt(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    transaction_hash: HexStr
    block_hash: HexStr
    block_number: HexInteger
    logs: list[Log]
    contract_address: Optional[HexStr]
    effective_gas_price: HexInteger
    cumulative_gas_used: HexInteger
    from_: HexAddress = Field(alias="from")
    gas_used: HexInteger
    logs_bloom: HexInteger
    status: Optional[HexInteger] = None
    to: Optional[HexAddress]
    transaction_index: HexInteger
    type: HexInteger

    def has_log(self, topic: HexStr):
        """Checks is a transaction has a topic in it"""
        t = bytes.fromhex(topic.replace("0x", ""))

        return t in BloomFilter(self.logs_bloom)



from eth_typing import HexStr
from pydantic import BaseModel

from ..basic import BlockReference, HexInteger


class TransactionRequest(BaseModel):
    tx_hash: HexStr


class GetTransactionByBlockHash(BaseModel):
    block_hash: HexStr
    index: HexInteger


class GetTransactionByBlockNumber(BaseModel):
    block_number: BlockReference
    index: HexInteger


class RawTransaction(BaseModel):
    signed_tx: HexStr



import pytest
from eth_rpc import Transaction
from eth_rpc.networks import Arbitrum, Ethereum


@pytest.mark.unit
def test_transaction_network() -> None:
    # Currently, block is a singleton
    tx = Transaction[Arbitrum]
    assert tx._network == Arbitrum

    tx2 = Transaction[Ethereum]
    assert tx2._network == Ethereum
    assert tx._network == Arbitrum



from typing import Generic, TypeVar

from eth_rpc.types import HexInteger, Network
from eth_rpc.utils import RPCModel
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel

T = TypeVar("T", bound=BaseModel)


class Log(RPCModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    transaction_hash: HexStr
    address: HexAddress
    block_hash: HexStr
    block_number: HexInteger
    data: HexStr
    log_index: HexInteger
    removed: bool
    topics: list[HexStr]
    transaction_index: HexInteger

    def __repr__(self):
        return f"<Log block={self.block_number} index={self.log_index}>"

    __str__ = __repr__


class EventData(BaseModel, Generic[T]):
    """
    This joins a log with it's transformed format (T).
    It needs to be in models because Models rely on Types.
    """

    name: str
    log: Log
    event: T
    network: Network

    @property
    def tx(self):
        # temporary while I remove event_data.tx
        return self.log

    def __hash__(self):
        # TODO: this could be more efficient
        return hash(
            f"{self.tx.transaction_hash}|{self.tx.block_number}|{self.tx.transaction_index}|{self.tx.log_index}"
        )

    def __repr__(self):
        return f"<name={self.name} log={self.log} {{{self.event}}}>>"

    __str__ = __repr__



from typing import Annotated

import pytest
from eth_rpc import FuncSignature
from eth_rpc.types import Name, map_type_to_str, primitives
from pydantic import BaseModel


class Inputs(BaseModel):
    x: primitives.uint256
    y: primitives.uint256
    z: list[primitives.address]
    a: primitives.address
    b: primitives.uint256
    c: list[primitives.address]
    d: primitives.uint256


class Outputs(BaseModel):
    x: Annotated[
        list[primitives.bytes32],
        Name("XX"),
    ]
    y: str


@pytest.mark.unit
def test_func_signature():
    func = FuncSignature[Inputs, Annotated[Outputs, Name("OUTPUT")]](
        name="swapExactTokensForETHWithPortfolios", alias="test_func"
    )
    # https://www.4byte.directory/signatures/?bytes4_signature=0x8b41f64c
    assert func.get_identifier() == "0x8b41f64c"
    assert func.get_inputs() == [
        "uint256",
        "uint256",
        "address[]",
        "address",
        "uint256",
        "address[]",
        "uint256",
    ]
    assert func.get_output() == ["bytes32[]", "string"]
    assert func._get_name(Outputs.__annotations__["x"]) == "XX"


@pytest.mark.unit
def test_map_to_str():
    assert map_type_to_str(primitives.address) == "address"
    assert map_type_to_str(str) == "string"
    assert map_type_to_str(list[str]) == "string[]"
    assert map_type_to_str(list[primitives.bytes32]) == "bytes32[]"



from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, PrivateAttr
from pydantic.alias_generators import to_camel

from ..types import Network as NetworkType
from .datetime import convert_datetime_to_iso_8601


class RPCModel(BaseModel):
    _metadata: dict[str, Any] = PrivateAttr(default_factory=dict)
    _network: Optional[NetworkType] = PrivateAttr(default=None)

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
        json_encoders={
            datetime: lambda v: convert_datetime_to_iso_8601(v),
        },
    )

    def set_network(self, network: NetworkType | None):
        self._network = network



from typing import Optional

from eth_rpc.types import HexInteger
from eth_typing import HexStr
from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel

from .access_list import AccessList


class BaseTransaction(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    hash: HexStr
    access_list: Optional[list[AccessList]] = None
    chain_id: Optional[HexInteger] = None
    from_: HexStr = Field(alias="from")
    gas: HexInteger
    gas_price: HexInteger
    max_fee_per_gas: Optional[HexInteger] = None
    max_priority_fee_per_gas: Optional[HexInteger] = None
    input: HexStr
    nonce: HexInteger
    r: HexStr
    s: HexStr
    v: HexInteger
    to: Optional[HexStr]
    type: Optional[HexInteger] = None
    value: HexInteger
    y_parity: HexInteger | None = None


class Transaction(BaseTransaction):
    block_hash: HexStr
    block_number: HexInteger
    transaction_index: HexInteger

    def get_block(self, with_tx_data: bool = False):
        """Load a block hash"""
        from ..block import Block

        return Block.load_by_hash(self.block_hash, with_tx_data=with_tx_data)


class PendingTransaction(BaseTransaction):
    block_hash: HexStr | None = None  # not set on pending transactions
    block_number: HexInteger | None = None  # not set on pending transactions
    transaction_index: HexInteger | None = None  # not set on pending transactions



from ._transport import (
    configure_rpc_from_env,
    get_current_network,
    set_alchemy_key,
    set_default_network,
    set_rpc_timeout,
    set_transport,
)
from .account import Account
from .block import Block
from .contract import Contract, ContractFunc, EthResponse, FuncSignature
from .contract.base import ProtocolBase
from .event import Event
from .log import Log
from .models import EventData
from .subscriber import EventSubscriber
from .transaction import Transaction, TransactionReceipt

__all__ = [
    "Account",
    "Block",
    "Contract",
    "ContractFunc",
    "EthResponse",
    "Event",
    "EventData",
    "EventSubscriber",
    "FuncSignature",
    "Log",
    "ProtocolBase",
    "Transaction",
    "TransactionReceipt",
    "configure_rpc_from_env",
    "get_current_network",
    "set_alchemy_key",
    "set_default_network",
    "set_transport",
    "set_rpc_timeout",
]



import asyncio
import json
from collections.abc import AsyncIterator
from typing import TypeVar

from eth_rpc.models import Log as LogModel
from eth_rpc.types import HexInt, LogsArgs, LogsParams
from eth_rpc.types import Network as NetworkType
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
from websockets.legacy.client import connect

from ._request import Request
from .block import Block
from .constants import DEFAULT_EVENT
from .types import RPCResponseModel

T = TypeVar("T")


class LogRPC(Request):
    def __class_getitem__(self, params):
        if isinstance(params, NetworkType):
            self._network = params
        return self

    def load_by_number(
        self,
        from_block: int | HexInt,
        to_block: int | HexInt | None = None,
    ) -> RPCResponseModel[LogsArgs, list["LogModel"]]:
        return RPCResponseModel(
            self._rpc().get_logs,
            LogsArgs(
                params=LogsParams(
                    from_block=HexInt(from_block),
                    to_block=HexInt(to_block or from_block + 1),
                )
            ),
        )

    @classmethod
    async def listen(
        cls,
        *,
        queue: asyncio.Queue["LogModel"],
        publish_logs: asyncio.Event = DEFAULT_EVENT,
    ):
        internal_queue: asyncio.Queue = asyncio.Queue()
        flush_queue: bool = True
        async for log in cls._listen():
            if publish_logs.is_set():
                if flush_queue:
                    while not internal_queue.empty():
                        staged_block = await internal_queue.get()
                        await queue.put(staged_block)
                    flush_queue = False
                await queue.put(log)
            else:
                await internal_queue.put(log)

    @classmethod
    async def _listen(cls):
        async for w3_connection in connect(
            cls._rpc().wss,
            ping_interval=60,
            ping_timeout=60,
            max_queue=10000,
        ):
            network = cls._network
            params = ["logs", {}]

            await w3_connection.send(
                json.dumps({"id": 1, "method": "eth_subscribe", "params": params})
            )
            subscription_response = await w3_connection.recv()
            if not (subscription_id := json.loads(subscription_response).get("result")):
                raise ValueError(subscription_response)
            assert subscription_id, "subscription failed to connect"

            while True:
                try:
                    raw_message = await asyncio.wait_for(
                        w3_connection.recv(), timeout=60
                    )
                    message_json = json.loads(raw_message)
                    result = message_json["params"]["result"]

                    yield Log(**result, network=network)
                except asyncio.exceptions.TimeoutError:
                    pass
                except (
                    ConnectionClosedOK,
                    ConnectionClosedError,
                    ConnectionResetError,
                    asyncio.exceptions.IncompleteReadError,  # TODO: should this be handled differently?
                ) as err:
                    print("Connection Error:", err)
                    # TODO: track the last block consumed, when you restart, confirm there's no gap in continuity
                    # w3_connection is an iterator, so make a new connection and continue listening
                    break

    async def subscribe_from(
        self,
        start_block: int | None = None,
        batch_size: int = 50,
    ) -> AsyncIterator[LogModel]:
        queue = asyncio.Queue[LogModel]()
        should_publish_logs = asyncio.Event()
        asyncio.create_task(
            self.listen(
                queue=queue,
                publish_logs=should_publish_logs,
            )
        )
        latest = await Block.get_number()
        if not start_block:
            start_block = latest
        assert start_block

        num = start_block
        while num <= latest:
            batch_end = min(num + batch_size, latest)
            for log in await self.load_by_number(num, batch_end):
                yield log
            num += batch_size

        should_publish_logs.set()
        while True:
            log = await queue.get()
            if log.block_number > latest:
                yield log


Log = LogRPC()



from collections.abc import Awaitable
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Generic, Literal, Optional, TypeVar, cast, overload

from eth_rpc.models import AccessListResponse
from eth_rpc.types import (
    BASIC_TYPES,
    BLOCK_STRINGS,
    CallWithBlockArgs,
    EthCallArgs,
    EthCallParams,
    HexInteger,
    MaybeAwaitable,
)
from eth_rpc.types import Network as NetworkType
from eth_rpc.types import RPCResponseModel
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel

from .._transport import _force_get_global_rpc
from ..block import Block
from ..constants import ADDRESS_ZERO
from ..rpc.core import RPC
from ..transaction import PreparedTransaction
from ..utils import run
from ..wallet import BaseWallet
from .eth_response import EthResponse
from .func_signature import FuncSignature

# from .contract import Contract

T = TypeVar(
    "T",
    bound=tuple
    | BaseModel
    | BASIC_TYPES
    | list[BASIC_TYPES]
    | tuple[BASIC_TYPES, ...]
    | HexAddress,
)
U = TypeVar("U")


@dataclass
class ContractFunc(Generic[T, U]):
    func: FuncSignature[T, U]
    contract: Any  # Contract
    data: HexStr = HexStr("0x")

    _network: NetworkType | None = None

    def __post_init__(self):
        self._network = self.contract._network

    @property
    def sync(self) -> "ContractFuncSync[T, U]":
        obj = deepcopy(self)
        obj.__class__ = ContractFuncSync
        obj = cast(ContractFuncSync, obj)
        return obj

    @property
    def address(self):
        return self.contract.address

    @property
    def name(self):
        return self.func.name

    @property
    def alias(self):
        return self.func.alias or self.func.name

    def _rpc(self) -> "RPC":
        """
        This uses the default network, unless a network has been provided, then immediately unsets the network.
        This makes it safe for async code.
        """
        from .._transport import _force_get_global_rpc

        network = self._network
        self._network = None
        response = _force_get_global_rpc(network)
        return response

    def __call__(
        self,
        *args: T,
    ):
        new_self = deepcopy(self)
        if len(args) == 0:
            new_self.data = self.func.encode_call(inputs=())  # type: ignore
        elif len(args) == 1:
            new_self.data = self.func.encode_call(inputs=args[0])
        else:
            # assumed you provided the args
            _args: T = cast(T, args)
            new_self.data = self.func.encode_call(inputs=_args)
        return new_self

    def encode(self):
        return bytes.fromhex(self.data[2:])

    def decode(self, result: bytes) -> U:
        return self.func.decode_result(HexStr(result.hex()))

    async def _estimate_gas(
        self,
        from_: Optional[HexAddress] = None,
        block_number: HexInteger | Literal["latest", "pending"] = "latest",
        sync: bool = False,
        buffer: float = 1.25,
    ) -> HexInteger:
        query: RPCResponseModel[CallWithBlockArgs, HexInteger] = RPCResponseModel(
            self._rpc().estimate_gas,
            CallWithBlockArgs(
                params=EthCallParams(
                    from_=from_,
                    to=self.address,
                    data=self.data,
                ),
                block_number=block_number,
            ),
        )
        if sync:
            return HexInteger(query.sync * buffer)
        return HexInteger((await query) * buffer)

    @overload
    def estimate_gas(
        self,
        *,
        sync: Literal[True],
        from_: Optional[HexAddress] = ...,
        block_number: HexInteger | Literal["latest", "pending"] = ...,
    ) -> HexInteger: ...

    @overload
    def estimate_gas(
        self,
        *,
        sync: Literal[False] = ...,
        from_: Optional[HexAddress] = ...,
        block_number: HexInteger | Literal["latest", "pending"] = ...,
    ) -> Awaitable[HexInteger]: ...

    def estimate_gas(
        self,
        *,
        from_: Optional[HexAddress] = None,
        block_number: HexInteger | Literal["latest", "pending"] = "latest",
        sync: bool = False,
    ) -> MaybeAwaitable[HexInteger]:
        return run(
            self._estimate_gas, from_=from_, block_number=block_number, sync=sync
        )

    async def _access_list(
        self,
        *,
        gas: int = 10000000,
        sender: HexAddress = ADDRESS_ZERO,
        block_number: int | None = None,
        sync: bool = False,
    ) -> AccessListResponse:
        if sync:
            return RPCResponseModel(
                self._rpc().create_access_list,
                CallWithBlockArgs(
                    params=EthCallParams(
                        from_=sender,
                        to=self.address,
                        data=self.data,
                        gas=HexInteger(gas),
                    ),
                    block_number=HexInteger(block_number) if block_number else None,
                ),
            ).sync
        return await RPCResponseModel(
            self._rpc().create_access_list,
            CallWithBlockArgs(
                params=EthCallParams(
                    from_=sender,
                    to=self.address,
                    data=self.data,
                    gas=HexInteger(gas),
                ),
                block_number=HexInteger(block_number) if block_number else None,
            ),
        )

    @overload
    def access_list(
        self,
        *,
        sync: Literal[True],
        gas: int = 10000000,
        sender: HexAddress = ADDRESS_ZERO,
        block_number: int | None = None,
    ) -> AccessListResponse: ...

    @overload
    def access_list(
        self,
        *,
        gas: int = 10000000,
        sender: HexAddress = ADDRESS_ZERO,
        block_number: int | None = None,
    ) -> Awaitable[AccessListResponse]: ...

    def access_list(
        self,
        *,
        gas: int = 10000000,
        sender: HexAddress = ADDRESS_ZERO,
        block_number: int | None = None,
        sync: bool = False,
    ) -> MaybeAwaitable[AccessListResponse]:
        return run(
            self._access_list,
            gas=gas,
            sender=sender,
            block_number=block_number,
            sync=sync,
        )

    async def _call(
        self,
        from_: Optional[HexAddress] = None,
        block_number: int | BLOCK_STRINGS = "latest",
        value: HexInteger | int = 0,
        state_diff: dict[HexAddress, Any] = {},
        sync: bool = True,
    ) -> EthResponse[T, U]:
        # TODO: as_dict feels useful to get back a dict with named fields, but it's odd
        #       to have a few different return types for the same function
        if self.contract.code_override:
            contract_state = state_diff.get(self.address, {})
            state_diff[self.address] = {
                "code": self.contract.code_override
            } | contract_state
        # TODO: make this handle sync or async
        if sync:
            response = self._rpc().eth_call.sync(
                EthCallArgs(
                    params=EthCallParams(
                        from_=from_,
                        to=self.address,
                        value=HexInteger(value),
                        data=self.data,
                    ),
                    block_number=(
                        HexInteger(block_number)
                        if isinstance(block_number, int)
                        else block_number
                    ),
                    state_override_set=state_diff,
                )
            )
        else:
            response = await self._rpc().eth_call(
                EthCallArgs(
                    params=EthCallParams(
                        from_=from_,
                        to=self.address,
                        value=HexInteger(value),
                        data=self.data,
                    ),
                    block_number=(
                        HexInteger(block_number)
                        if isinstance(block_number, int)
                        else block_number
                    ),
                    state_override_set=state_diff,
                )
            )
        return EthResponse[T, U](
            result=response,
            func=self.func,
        )

    @overload
    def call(
        self,
        *,
        sync: Literal[True],
        from_: Optional[HexAddress] = ...,
        block_number: int | BLOCK_STRINGS = ...,
        value: HexInteger | int = ...,
        state_diff: dict[HexAddress, Any] = ...,
    ) -> EthResponse[T, U]: ...

    @overload
    def call(
        self,
        *,
        sync: Literal[False],
        from_: Optional[HexAddress] = ...,
        block_number: int | BLOCK_STRINGS = ...,
        value: HexInteger | int = ...,
        state_diff: dict[HexAddress, Any] = ...,
    ) -> Awaitable[EthResponse[T, U]]: ...

    @overload
    def call(
        self,
        *,
        from_: Optional[HexAddress] = ...,
        block_number: int | BLOCK_STRINGS = ...,
        value: HexInteger | int = ...,
        state_diff: dict[HexAddress, Any] = ...,
    ) -> Awaitable[EthResponse[T, U]]: ...

    @overload
    def call(
        self,
        *,
        sync: bool,
        from_: Optional[HexAddress] = ...,
        block_number: int | BLOCK_STRINGS = ...,
        value: HexInteger | int = ...,
        state_diff: dict[HexAddress, Any] = ...,
    ) -> MaybeAwaitable[EthResponse[T, U]]: ...

    def call(
        self,
        *,
        from_: Optional[HexAddress] = None,
        block_number: int | BLOCK_STRINGS = "latest",
        value: HexInteger | int = 0,
        state_diff: dict[HexAddress, Any] = {},
        sync: bool = False,
    ) -> MaybeAwaitable[EthResponse[T, U]]:
        return run(self._call, from_, block_number, value, state_diff, sync=sync)

    async def _get(
        self,
        from_: Optional[HexAddress] = None,
        block_number: int | BLOCK_STRINGS = "latest",
        value: HexInteger | int = 0,
        state_diff: dict[HexAddress, Any] = {},
        sync: bool = False,
    ) -> U:
        """shorthand for call().decode()"""
        return (
            await self._call(
                from_=from_,
                block_number=block_number,
                value=value,
                state_diff=state_diff,
                sync=sync,
            )
        ).decode()

    @overload
    def get(
        self,
        *,
        sync: Literal[True],
        from_: Optional[HexAddress] = ...,
        block_number: int | BLOCK_STRINGS = ...,
        value: HexInteger | int = ...,
        state_diff: dict[HexAddress, Any] = ...,
    ) -> U: ...

    @overload
    def get(
        self,
        *,
        sync: Literal[False] = ...,
        from_: Optional[HexAddress] = ...,
        block_number: int | BLOCK_STRINGS = ...,
        value: HexInteger | int = ...,
        state_diff: dict[HexAddress, Any] = ...,
    ) -> Awaitable[U]: ...

    def get(
        self,
        *,
        from_: Optional[HexAddress] = None,
        block_number: int | BLOCK_STRINGS = "latest",
        value: HexInteger | int = 0,
        state_diff: dict[HexAddress, Any] = {},
        sync: bool = False,
    ) -> MaybeAwaitable[U]:
        return run(
            self._get,
            from_=from_,
            block_number=block_number,
            value=value,
            state_diff=state_diff,
            sync=sync,
        )

    async def _prepare(
        self,
        wallet: "BaseWallet",
        *,
        nonce: Optional[int] = None,
        value: int = 0,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
        use_access_list: bool = True,
        sync: bool = False,
        buffer: float = 1.25,
    ) -> PreparedTransaction:
        gas = await self._estimate_gas(from_=wallet.address, sync=sync, buffer=buffer)
        if use_access_list:
            if sync:
                access_list_response = self.access_list(
                    sender=wallet.address, sync=True
                )
            else:
                access_list_response = await self.access_list(sender=wallet.address)
            access_list = access_list_response.access_list
        else:
            access_list = None
        rpc = _force_get_global_rpc()
        if sync:
            chain_id = rpc.chain_id.sync()
        else:
            chain_id = await rpc.chain_id()

        if sync:
            max_priority_fee_per_gas = (
                max_priority_fee_per_gas or Block.priority_fee().sync
            )
            base_fee_per_gas = Block.pending().sync.base_fee_per_gas
        else:
            max_priority_fee_per_gas = (
                max_priority_fee_per_gas or await Block.priority_fee()
            )
            # TODO: fix pending()
            base_fee_per_gas = (await Block.pending()).base_fee_per_gas

        assert base_fee_per_gas, "block is earlier than London Hard Fork"
        max_fee_per_gas = max_fee_per_gas or (
            2 * base_fee_per_gas + max_priority_fee_per_gas
        )

        if sync:
            return PreparedTransaction(
                data=self.data,
                to=self.address,
                gas=HexInteger(gas),
                max_fee_per_gas=max_fee_per_gas,
                max_priority_fee_per_gas=max_priority_fee_per_gas,
                nonce=nonce or wallet.get_nonce().sync,
                value=value,
                access_list=access_list,
                chain_id=chain_id,
            )
        return PreparedTransaction(
            data=self.data,
            to=self.address,
            gas=HexInteger(gas),
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            nonce=nonce or await wallet.get_nonce(),
            value=value,
            access_list=access_list,
            chain_id=chain_id,
        )

    @overload
    def prepare(
        self,
        wallet: "BaseWallet",
        *,
        sync: Literal[True],
        nonce: Optional[int] = ...,
        value: int = ...,
        max_fee_per_gas: Optional[int] = ...,
        max_priority_fee_per_gas: Optional[int] = ...,
        use_access_list: bool = ...,
    ) -> PreparedTransaction: ...

    @overload
    def prepare(
        self,
        wallet: "BaseWallet",
        *,
        nonce: Optional[int] = ...,
        value: int = ...,
        max_fee_per_gas: Optional[int] = ...,
        max_priority_fee_per_gas: Optional[int] = ...,
        use_access_list: bool = ...,
    ) -> Awaitable[PreparedTransaction]: ...

    def prepare(
        self,
        wallet: "BaseWallet",
        *,
        nonce: Optional[int] = None,
        value: int = 0,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
        use_access_list: bool = True,
        sync: bool = False,
    ) -> MaybeAwaitable[PreparedTransaction]:
        return run(
            self._prepare,
            wallet,
            nonce=nonce,
            value=value,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            use_access_list=use_access_list,
            sync=sync,
        )

    async def _execute(
        self,
        wallet: "BaseWallet",
        *,
        nonce: Optional[int] = None,
        value: int = 0,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
        use_access_list: bool = True,
        sync: bool = True,
    ) -> HexStr:
        if sync is True:
            prepared_tx = self.prepare(
                wallet,
                nonce=nonce,
                value=value,
                max_fee_per_gas=max_fee_per_gas,
                max_priority_fee_per_gas=max_priority_fee_per_gas,
                use_access_list=use_access_list,
                sync=True,
            )
        else:
            prepared_tx = await self.prepare(
                wallet,
                nonce=nonce,
                value=value,
                max_fee_per_gas=max_fee_per_gas,
                max_priority_fee_per_gas=max_priority_fee_per_gas,
                use_access_list=use_access_list,
            )
        signed_tx = wallet.sign_transaction(prepared_tx)
        if sync:
            return wallet.send_raw_transaction(
                HexStr("0x" + signed_tx.raw_transaction)
            ).sync
        return await wallet.send_raw_transaction(
            HexStr("0x" + signed_tx.raw_transaction)
        )

    @overload
    def execute(
        self,
        wallet: "BaseWallet",
        *,
        sync: Literal[True],
        nonce: Optional[int] = ...,
        value: int = ...,
        max_fee_per_gas: Optional[int] = ...,
        max_priority_fee_per_gas: Optional[int] = ...,
        use_access_list: bool = ...,
    ) -> HexStr: ...

    @overload
    def execute(
        self,
        wallet: "BaseWallet",
        *,
        nonce: Optional[int] = ...,
        value: int = ...,
        max_fee_per_gas: Optional[int] = ...,
        max_priority_fee_per_gas: Optional[int] = ...,
        use_access_list: bool = ...,
    ) -> Awaitable[HexStr]: ...

    def execute(
        self,
        wallet: "BaseWallet",
        *,
        nonce: Optional[int] = None,
        value: int = 0,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
        use_access_list: bool = True,
        sync: bool = True,
    ) -> MaybeAwaitable[HexStr]:
        return run(
            self._execute,
            wallet,
            nonce=nonce,
            value=value,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            use_access_list=use_access_list,
            sync=sync,
        )

    def _encode_block_number(
        self, block_number: int | Literal["latest", "pending"] = "pending"
    ):
        if isinstance(block_number, int):
            return hex(block_number)
        return block_number


class ContractFuncSync(ContractFunc[T, U]):
    SYNC: Literal[True] = True

    def estimate_gas(  # type: ignore
        self,
        *,
        from_: Optional[HexAddress] = None,
        block_number: HexInteger | Literal["latest", "pending"] = "latest",
    ) -> HexInteger:
        return super().estimate_gas(
            from_=from_, block_number=block_number, sync=self.SYNC
        )

    def access_list(  # type: ignore
        self,
        *,
        gas: int = 10000000,
        sender: HexAddress = ADDRESS_ZERO,
        block_number: int | None = None,
    ) -> AccessListResponse:
        return super().access_list(
            gas=gas,
            sender=sender,
            block_number=block_number,
            sync=self.SYNC,
        )

    def call(  # type: ignore
        self,
        *,
        from_: Optional[HexAddress] = None,
        block_number: int | BLOCK_STRINGS = "latest",
        value: HexInteger | int = 0,
        state_diff: dict[HexAddress, Any] = {},
    ) -> EthResponse[T, U]:
        return super().call(
            from_=from_,
            block_number=block_number,
            value=value,
            state_diff=state_diff,
            sync=self.SYNC,
        )

    def get(  # type: ignore
        self,
        *,
        from_: Optional[HexAddress] = None,
        block_number: int | BLOCK_STRINGS = "latest",
        value: HexInteger | int = 0,
        state_diff: dict[HexAddress, Any] = {},
    ) -> U:
        return super().get(
            from_=from_,
            block_number=block_number,
            value=value,
            state_diff=state_diff,
            sync=self.SYNC,
        )

    def prepare(  # type: ignore
        self,
        wallet: "BaseWallet",
        *,
        nonce: Optional[int] = None,
        value: int = 0,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
        use_access_list: bool = True,
    ) -> PreparedTransaction:
        return super().prepare(
            wallet,
            nonce=nonce,
            value=value,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            use_access_list=use_access_list,
            sync=self.SYNC,
        )

    def execute(  # type: ignore
        self,
        wallet: BaseWallet,
        *,
        nonce: Optional[int] = None,
        value: int = 0,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
        use_access_list: bool = True,
    ) -> HexStr:
        return super().execute(
            wallet,
            nonce=nonce,
            value=value,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            use_access_list=use_access_list,
            sync=self.SYNC,
        )



from typing import Awaitable, Callable, Generic, ParamSpec, TypeVar

from eth_rpc.types import HexInt, NoArgs
from eth_typing import HexStr
from pydantic import BaseModel

from .base import RPCMethodBase

Params = TypeVar(
    "Params", bound=BaseModel | HexInt | HexStr | list[HexInt | HexStr] | NoArgs
)
Response = TypeVar("Response")
P = ParamSpec("P")


class RPCMethod(RPCMethodBase[Params, Response], Generic[Params, Response]):
    def __call__(self, *params: Params) -> Awaitable[Response]:
        return self.call_async(*params)

    @property
    def sync(self) -> Callable[..., Response]:
        return self.call_sync



from eth_hash.auto import keccak
from eth_rpc.contract import ContractFunc, ProtocolBase
from eth_rpc.networks import Ethereum
from eth_rpc.types import METHOD, BlockReference, Network, primitives


class ENSRegistryWithFallback(ProtocolBase):
    resolver: ContractFunc[
        primitives.bytes32,
        primitives.address,
    ] = METHOD


class Resolver(ProtocolBase):
    addr: ContractFunc[
        primitives.bytes32,
        primitives.address,
    ] = METHOD


async def lookup_addr(
    name: str, block_number: BlockReference = "latest", network: Network = Ethereum
):
    hashed_name = hash_ens_name(name)

    registry = ENSRegistryWithFallback(address=network.ens_registry)
    resolver_address = await registry.resolver(hashed_name).get(
        block_number=block_number
    )

    resolver = Resolver(address=resolver_address)
    return await resolver.addr(hashed_name).get(block_number=block_number)


def hash_ens_name(name: str):
    node = b"\0" * 32
    labels = name.split(".")
    for label in reversed(labels):
        labelhash = keccak(label.encode("utf-8"))
        assert isinstance(labelhash, bytes)
        assert isinstance(node, bytes)
        node = keccak(node + labelhash)
    return node



from collections.abc import AsyncIterator

from eth_rpc.models import EventData
from eth_rpc.utils import acombine
from pydantic import BaseModel

from ..event import Event


class EventCombinator(BaseModel):
    events: list[Event]

    async def subscribe(self) -> AsyncIterator[EventData]:
        async for event in acombine(*[event.subscribe() for event in self.events]):
            yield event

    async def backfill(
        self, start_block: int, end_block: int, step_size: int | None = None
    ) -> AsyncIterator[EventData]:
        for event in acombine(
            *[
                event.backfill(
                    start_block=start_block,
                    end_block=end_block,
                    step_size=step_size,
                )
                for event in self.events
            ]
        ):
            yield event



from eth_abi import encode
from eth_typing import ChecksumAddress, HexAddress, HexStr
from eth_utils import to_checksum_address


def address_to_topic(address: str) -> HexStr:
    return HexStr("0x{:064x}".format(int(address, 16)))


def address_to_bytes32(address: HexAddress | ChecksumAddress | str):
    return encode(["bytes32"], [bytes.fromhex(address.lstrip("0x"))])


def to_checksum(address: str | HexAddress | ChecksumAddress) -> ChecksumAddress:
    return ChecksumAddress(HexAddress(HexStr(to_checksum_address(address))))



from dataclasses import dataclass


class Indexed:
    """Used to annotate when an event field is indexed"""


@dataclass(eq=True, frozen=True)
class Name:
    """Used to annotate a name for variables different than its declared name"""

    value: str



from types import GenericAlias
from typing import Annotated, Any, Literal, get_args

from eth_typing import ChecksumAddress, HexAddress, HexStr
from pydantic import (
    BaseModel,
    GetCoreSchemaHandler,
    SerializerFunctionWrapHandler,
    ValidationInfo,
)
from pydantic.functional_serializers import WrapSerializer
from pydantic_core import core_schema

from . import primitives

ALL_PRIMITIVES = [x for x in dir(primitives) if x[0].islower() and x[0] != "_"]
BLOCK_STRINGS = Literal["latest", "earliest", "pending", "safe", "finalized"]


class Empty(BaseModel): ...


class HexInt(int):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            function=cls.validate,
            schema=handler(int | str),
            field_name=handler.field_name,
        )

    @classmethod
    def validate(cls, v, r: ValidationInfo):
        if isinstance(v, str):
            return int(v, 16)
        return int(v)


def hex_int_wrap(v: Any, nxt: SerializerFunctionWrapHandler) -> str:
    if v in get_args(BLOCK_STRINGS):
        return v
    return hex(v)


# A Custom Integer type that loads from hex string and serializes to hex string
# TODO: make this a BlockNumber type and support block strings too
HexInteger = Annotated[HexInt, WrapSerializer(hex_int_wrap)]
BlockReference = HexInteger | BLOCK_STRINGS


def map_type_to_str(type_: Any) -> str:
    """Convert a type to its solidity compatible format"""
    mapping = {
        str: "string",
        int: "uint256",  # defaults to uint
        HexAddress: "address",
        ChecksumAddress: "address",
        HexStr: "bytes",
    }
    if type_ in mapping:
        return mapping[type_]
    if str(type_) in ALL_PRIMITIVES:
        return str(type_)
    # handle list type
    if isinstance(type_, GenericAlias):
        (arg,) = get_args(type_)
        return f"{map_type_to_str(arg)}[]"
    return type_.__name__



from eth_rpc.models import Account as AccountModel
from eth_rpc.types import BLOCK_STRINGS
from eth_rpc.types.args import GetAccountArgs
from eth_typing import HexAddress

from ._request import Request
from .types import HexInteger, RPCResponseModel


class AccountRPC(Request):
    def get_balance(
        self, address: HexAddress, block_number: int | BLOCK_STRINGS = "latest"
    ) -> RPCResponseModel[GetAccountArgs, HexInteger]:
        return RPCResponseModel(
            self._rpc().get_balance,
            GetAccountArgs(
                address=address,
                block_number=(
                    HexInteger(block_number)
                    if isinstance(block_number, int)
                    else block_number
                ),
            ),
        )

    def get_account(
        self, address: HexAddress, block_number: int | BLOCK_STRINGS = "latest"
    ) -> RPCResponseModel[GetAccountArgs, AccountModel]:
        return RPCResponseModel(
            self._rpc().get_account,
            GetAccountArgs(
                address=address,
                block_number=(
                    HexInteger(block_number)
                    if isinstance(block_number, int)
                    else block_number
                ),
            ),
        )


Account = AccountRPC()



from typing import Any, Literal

from pydantic import BaseModel

from .eth_call import CallWithBlockArgs


class TracerConfig(BaseModel):
    tracer: Literal["callTracer", "prestateTracer"] = "prestateTracer"
    only_top_call: bool | None = None

    def model_dump(self, exclude_none=True, **kwargs) -> dict[str, Any]:
        return super().model_dump(exclude_none=exclude_none, **kwargs)


class TraceArgs(CallWithBlockArgs):
    tracer_config: TracerConfig | None = TracerConfig()



import os

import pytest
from eth_rpc import Block, set_alchemy_key
from eth_rpc.networks import Arbitrum, Ethereum


@pytest.mark.unit
def test_block_network() -> None:
    # Currently, block is a singleton
    block = Block[Arbitrum]
    assert block._network == Arbitrum

    block2 = Block[Ethereum]
    assert Block._network is None
    assert block._network == Arbitrum
    assert block2._network == Ethereum


@pytest.mark.asyncio(scope="session")
async def test_block_methods() -> None:
    set_alchemy_key(os.environ["ALCHEMY_KEY"])
    assert Block[Arbitrum]._network == Arbitrum
    assert Block[Ethereum]._network == Ethereum

    ten_million_eth = await Block[Ethereum].load_by_number(10_000_000)
    assert ten_million_eth.number == 10_000_000
    assert len(ten_million_eth.transactions) == 103
    TEN_MILLION_BLOCK_HASH_ETH = (
        "0xaa20f7bde5be60603f11a45fc4923aab7552be775403fc00c2e6b805e6297dbe"
    )
    assert ten_million_eth.hash == TEN_MILLION_BLOCK_HASH_ETH

    ten_million_arb = await Block[Arbitrum].load_by_number(10_000_000)
    TEN_MILLION_BLOCK_HASH_ARB = (
        "0xed03835298dc4bfbb5da9c6d26aa8c8bf96bb58bcd71ea868f3b356de51ac35a"
    )
    assert ten_million_arb.hash == TEN_MILLION_BLOCK_HASH_ARB

    assert (await Block[Ethereum].load_by_hash(ten_million_eth.hash)) == ten_million_eth
    assert (await Block[Arbitrum].load_by_hash(ten_million_arb.hash)) == ten_million_arb

    fee_history = await Block[Ethereum].fee_history()
    assert fee_history



from .arbitrum import Arbitrum
from .base import Base
from .ethereum import Ethereum

Networks = {
    1: Ethereum,
    8453: Base,
    42161: Arbitrum,
}


def get_network_by_chain_id(chain_id):
    return Networks.get(chain_id)


__all__ = [
    "Arbitrum",
    "Base",
    "Ethereum",
    "Networks",
]



from eth_rpc.types import HexInteger
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class FeeHistory(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    oldest_block: HexInteger
    base_fee_per_gas: list[HexInteger]
    base_fee_per_blob_gas: list[HexInteger] | None = None
    gas_used_ratio: list[float]
    reward: list[list[HexInteger]]

    def mean_fee(self, buffer=1):
        """gas_price"""
        return int(buffer * sum(self.base_fee_per_gas) / len(self.base_fee_per_gas))



from collections.abc import AsyncIterator, Iterator
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from eth_rpc.models import EventData


def sort_key(row: "EventData"):
    return (row.tx.block_number, row.tx.transaction_index, row.tx.log_index)


def combine(*args: Iterator["EventData"]):
    generators: list[Iterator["EventData"]] = list(args)
    values = [next(generator) for generator in generators]
    dedupe = {}
    while len(generators):
        min_index = values.index(min(values, key=sort_key))
        min_value = values[min_index]
        try:
            values[min_index] = next(generators[min_index])
        except StopAsyncIteration:
            generators.pop(min_index)
            values.pop(min_index)
        if (key := sort_key(min_value)) not in dedupe:
            dedupe[key] = True
            yield min_value


async def acombine(*args: AsyncIterator["EventData"]):
    generators: list[AsyncIterator["EventData"]] = list(args)

    values = []
    for generator in generators:
        try:
            value = await anext(generator)
            values.append(value)
        except StopAsyncIteration:
            continue  # Skip this generator if it's already exhausted

    dedupe = {}
    while len(generators):
        if not generators or not values:
            break
        min_index = values.index(min(values, key=sort_key))
        min_value = values[min_index]
        try:
            values[min_index] = await anext(generators[min_index])
        except StopAsyncIteration:
            generators.pop(min_index)
            values.pop(min_index)
        if (key := sort_key(min_value)) not in dedupe:
            dedupe[key] = True
            yield min_value


def ordered_iterator(events: list["EventData"]) -> Iterator["EventData"]:
    for event in sorted(
        events,
        key=lambda x: (x.tx.block_number, x.tx.transaction_index, x.tx.log_index),
    ):
        yield event



from types import GenericAlias
from typing import get_args

from eth_abi import encode
from eth_rpc.utils import convert_base_model
from eth_typing import ChecksumAddress, HexAddress, HexStr
from pydantic import BaseModel

from . import primitives

ALL_PRIMITIVES = [x for x in dir(primitives) if x[0].islower() and x[0] != "_"]


class Struct(BaseModel):
    """
    This is a type used to denote a solidity struct.  It is necessary
    to set this so the encoder knows to treat it as a single struct,
    and not a list of fields.

    For example, (uint8,bytes32) will encode differently from ((uint8,bytes32))

    One is a field with two arguments, the other is a field with a single argument,
        that is a struct with two fields
    """

    @staticmethod
    def transform(type_):
        mapping = {
            str: "string",
            int: "uint256",  # defaults to uint
            HexAddress: "address",
            ChecksumAddress: "address",
            HexStr: "bytes",
        }
        if type_ in mapping:
            return mapping[type_]
        if str(type_) in ALL_PRIMITIVES:
            return str(type_)
        # handle list type
        if isinstance(type_, GenericAlias):
            (arg,) = get_args(type_)
            return f"{arg.__name__}[]"
        return type_.__name__

    def to_bytes(self):
        # TODO: this should be able to handle recursive types, it will fail if the depth is more than 2
        types = ",".join(convert_base_model(self.__class__))
        values = []
        for _, value in self:
            # this generally works, but should be revisited
            if (
                isinstance(value, list)
                and len(value) > 0
                and isinstance(value[0], Struct)
            ):
                value = [tuple(v.model_dump().values()) for v in value]
            elif isinstance(value, Struct):
                value = tuple(value.model_dump().values())
            values.append(value)
        return encode([f"({types})"], [tuple(values)])



import asyncio
import json
import zlib
from collections.abc import AsyncIterator
from contextvars import ContextVar
from datetime import datetime, timezone
from typing import Literal

from eth_rpc.models import Block as BlockModel
from eth_rpc.models import FeeHistory
from eth_rpc.types.args import (
    BlockNumberArg,
    FeeHistoryArgs,
    GetBlockByHashArgs,
    GetBlockByNumberArgs,
)
from eth_typing import HexStr
from typing_extensions import Self
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
from websockets.legacy.client import connect

from ._request import Request
from .constants import DEFAULT_EVENT
from .models import Transaction as TransactionModel
from .types import BLOCK_STRINGS, BlockReference, HexInteger, NoArgs, RPCResponseModel

SUBSCRIPTION_TYPE = Literal["newHeads", "newPendingTransactions"]
DEFAULT_CONTEXT = ContextVar[int]("DEFAULT_CONTEXT")
DEFAULT_CONTEXT.set(0)


class BlockRPC(Request):
    def priority_fee(self) -> RPCResponseModel[NoArgs, HexInteger]:
        return RPCResponseModel(
            self._rpc().max_priority_fee_per_gas,
        )

    def fee_history(
        self,
        block_count: int = 4,
        lower_percentile: int = 25,
        upper_percentile: int = 75,
        block_number: BlockReference = "latest",
    ) -> RPCResponseModel[FeeHistoryArgs, FeeHistory]:
        return RPCResponseModel(
            self._rpc().fee_history,
            FeeHistoryArgs(
                block_count=block_count,
                block_number=block_number,
                percentiles=[
                    lower_percentile,
                    upper_percentile,
                ],
            ),
        )

    def get_block_transaction_count(
        self, block_number: HexInteger
    ) -> RPCResponseModel[BlockNumberArg, int]:
        return RPCResponseModel(
            self._rpc().get_block_tx_count_by_number,
            BlockNumberArg(
                block_number=block_number,
            ),
        )

    def load_by_number(
        self,
        block_number: int | HexInteger | BLOCK_STRINGS,
        with_tx_data: bool = False,
    ) -> RPCResponseModel[GetBlockByNumberArgs, BlockModel]:
        return RPCResponseModel(
            self._rpc().get_block_by_number,
            GetBlockByNumberArgs(
                block_number=(
                    HexInteger(block_number)
                    if isinstance(block_number, int)
                    else block_number
                ),
                with_tx_data=with_tx_data,
            ),
        )

    def get_number(self) -> RPCResponseModel[NoArgs, HexInteger]:
        return RPCResponseModel(
            self._rpc().block_number,
        )

    @classmethod
    async def load_by_datetime(
        cls,
        when: datetime,
        low: int | None = None,
        high: int | None = None,
        apprx_block_time=12,
    ) -> Self:
        """
        Searches for a block, finding the first block before a datetime.
        Recursively searches, using low and high as the boundaries for the binary search.
        """
        Network = cls._network
        if not when.tzinfo:
            when = when.replace(tzinfo=timezone.utc)
        if not (low and high):
            now = datetime.now(timezone.utc)
            diff = now - when
            day_diff = diff.days
            seconds_diff = day_diff * 24 * 3600
            block_number = await cls[Network].get_number()  # type: ignore
            if not low:
                low = int(
                    max(block_number - (seconds_diff / (apprx_block_time * 0.8)), 0)
                )
            if not high:
                high = int(
                    min(
                        block_number - (seconds_diff / (apprx_block_time * 1.2)),
                        block_number,
                    )
                )
        if when < datetime(
            year=2015,
            month=7,
            day=30,
            hour=3,
            minute=26,
            second=13,
            tzinfo=timezone.utc,
        ):
            raise ValueError("Block before genesis")

        if high > low:
            mid = (high + low) // 2
            mid_block = await cls[Network].load_by_number(mid)  # type: ignore

            if mid_block.timestamp == when:
                return mid_block
            elif mid_block.timestamp > when:
                return await cls[Network].load_by_datetime(when, low, mid - 1)  # type: ignore
            else:
                return await cls[Network].load_by_datetime(when, mid + 1, high)  # type: ignore
        return await cls[Network].load_by_number(high)  # type: ignore

    def latest(
        self, with_tx_data: bool = False
    ) -> RPCResponseModel[GetBlockByNumberArgs, "BlockModel"]:
        return self.load_by_number("latest", with_tx_data=with_tx_data)

    def pending(
        self, with_tx_data: bool = False
    ) -> RPCResponseModel[GetBlockByNumberArgs, "BlockModel"]:
        return self.load_by_number("pending", with_tx_data=with_tx_data)

    def load_by_hash(
        self, block_hash: HexStr, with_tx_data: bool = False
    ) -> RPCResponseModel[GetBlockByHashArgs, "BlockModel"]:
        return RPCResponseModel(
            self._rpc().get_block_by_hash,
            GetBlockByHashArgs(
                block_hash=block_hash,
                with_tx_data=with_tx_data,
            ),
        )

    @staticmethod
    def _to_hex(number: int | str) -> HexStr:
        if isinstance(number, int):
            return HexStr(hex(number))
        return HexStr(number)

    async def subscribe_from(
        self,
        start_block: int | None = None,
        with_tx_data: bool = True,
    ) -> AsyncIterator["BlockModel"]:
        queue = asyncio.Queue[BlockModel]()
        should_publish_blocks = asyncio.Event()
        asyncio.create_task(
            self.listen(
                queue=queue,
                publish_blocks=should_publish_blocks,
                with_tx_data=with_tx_data,
            )
        )
        latest = await Block.latest()
        if not start_block:
            start_block = latest.number
        assert start_block

        # NOTE: you pull latest twice because there can be a backfill while you're populating
        for num in range(start_block, latest.number + 1):
            yield await Block.load_by_number(num, with_tx_data=with_tx_data)

        should_publish_blocks.set()
        while True:
            block = await queue.get()
            if block.number > latest.number:
                yield block

    async def listen(
        self,
        *,
        # TODO: typehinting this is tricky because the type of the Queue is conditional based on the subscription type
        queue: asyncio.Queue,
        publish_blocks: asyncio.Event = DEFAULT_EVENT,
        with_tx_data: bool = True,
        subscription_type: SUBSCRIPTION_TYPE = "newHeads",
    ):
        internal_queue: asyncio.Queue = asyncio.Queue()
        flush_queue: bool = True
        async for block in self._listen(
            with_tx_data=with_tx_data, subscription_type=subscription_type
        ):
            if publish_blocks.is_set():
                if flush_queue:
                    while not internal_queue.empty():
                        staged_block = await internal_queue.get()
                        await queue.put(staged_block)
                    flush_queue = False
                await queue.put(block)
            else:
                await internal_queue.put(block)

    async def _listen(  # noqa: C901
        self,
        with_tx_data: bool = True,
        subscription_type: SUBSCRIPTION_TYPE = "newHeads",
    ):
        async for w3_connection in connect(
            self._rpc().wss,
            ping_interval=60,
            ping_timeout=60,
            max_queue=10000,
        ):
            params: list
            params = [subscription_type]

            await w3_connection.send(
                json.dumps({"id": 1, "method": "eth_subscribe", "params": params})
            )
            subscription_response = await w3_connection.recv()
            if not (subscription_id := json.loads(subscription_response).get("result")):
                raise ValueError(subscription_response)
            assert subscription_id, "subscription failed to connect"

            while True:
                try:
                    raw_message = await asyncio.wait_for(
                        w3_connection.recv(), timeout=60
                    )
                    message_json = json.loads(raw_message)
                    result = message_json["params"]["result"]

                    if subscription_type == "newHeads":
                        if not with_tx_data:
                            block_ = BlockModel(**result)
                        else:
                            block_ = await self.load_by_hash(
                                result["hash"], with_tx_data=with_tx_data
                            )
                        yield block_
                    else:
                        yield TransactionModel(**result)
                except asyncio.exceptions.TimeoutError:
                    pass
                except (
                    ConnectionClosedOK,
                    ConnectionClosedError,
                    ConnectionResetError,
                    asyncio.exceptions.IncompleteReadError,  # TODO: should this be handled differently?
                ) as err:
                    print("Connection Error:", err)
                    # TODO: track the last block consumed, when you restart, confirm there's no gap in continuity
                    # w3_connection is an iterator, so make a new connection and continue listening
                    break

    async def convert(self, block_value: BLOCK_STRINGS | int) -> int:
        if isinstance(block_value, int):
            return block_value
        block = await self.load_by_number(block_value)
        return block.number

    @classmethod
    def decompress(cls, raw_bytes: bytes) -> "BlockModel":
        """Convert gzip compressed block to a Block"""
        return BlockModel.model_validate_json(zlib.decompress(raw_bytes))


Block = BlockRPC()



from copy import deepcopy
from typing import TYPE_CHECKING, ClassVar

from eth_rpc.types.args import EthCallParams, TraceArgs
from pydantic import BaseModel, PrivateAttr

from .types import BlockReference, Network

if TYPE_CHECKING:
    from .rpc.core import RPC


class Request(BaseModel):
    _network_: ClassVar[Network | None] = PrivateAttr(default=None)  # type: ignore
    _network: Network | None = PrivateAttr(default=None)  # type: ignore

    def model_post_init(self, __context):
        network = self.__class__._network_
        self._network = network

    def __class_getitem__(cls, params):
        if isinstance(params, Network):
            cls._network_ = params
        else:
            try:
                return super().__class_getitem__(params)
            except AttributeError:
                pass
        return cls

    def __getitem__(self, params):
        if isinstance(params, Network) and self._network != params:
            self = deepcopy(self)
            self._network = params
        return self

    def _rpc(self) -> "RPC":
        """
        This uses the default network, unless a network has been provided
        """
        from ._transport import _force_get_global_rpc

        network = self._network
        response = _force_get_global_rpc(network)
        return response

    def _get_debug_tracecall(
        self, address, data, block_number: BlockReference = "latest"
    ):
        return self._rpc().debug_tracecall(
            TraceArgs(
                params=EthCallParams(
                    to=address,
                    data=data,
                ),
                block_number=block_number,
            )
        )



from collections.abc import Generator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from eth_typing import HexAddress, HexStr
from pydantic import BaseModel

from ..types import BASIC_TYPES

if TYPE_CHECKING:
    from .function import FuncSignature

T = TypeVar(
    "T",
    bound=tuple
    | BaseModel
    | BASIC_TYPES
    | list[BASIC_TYPES]
    | tuple[BASIC_TYPES, ...]
    | HexAddress,
)
U = TypeVar("U")


@dataclass
class EthResponse(Generic[T, U]):
    func: "FuncSignature[T, U]"
    result: HexStr

    @property
    def raw(self):
        return self.result

    def decode(self) -> U:
        return self.func.decode_result(self.result)

    def as_dict(self, **kwargs) -> dict[str, Any]:
        decoded_result = self.func.decode_result(self.result)
        # NOTE: tuple is redundant here
        if isinstance(decoded_result, tuple):
            return dict(
                zip(
                    [
                        (output_name or str(idx))
                        for idx, output_name in enumerate(self.func.get_output_name())
                    ],
                    decoded_result,
                )
            )
        elif isinstance(decoded_result, BaseModel):
            return decoded_result.model_dump(**kwargs)
        return {self.func.get_output_name()[0]: decoded_result}

    def __await__(self) -> "Generator[Any, None, EthResponse[T, U]]":
        async def foo():
            return self

        return foo().__await__()



import os
from typing import Annotated

import pytest
from eth_rpc import Event, EventData, EventSubscriber, set_alchemy_key
from eth_rpc.types import Indexed, primitives
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel
from typing_extensions import assert_type


class TransferEventType(BaseModel):
    sender: Annotated[primitives.address, Indexed]
    recipient: Annotated[primitives.address, Indexed]
    amount: primitives.uint256


class ApprovalEventType(BaseModel):
    owner: Annotated[primitives.address, Indexed]
    spender: Annotated[primitives.address, Indexed]
    value: primitives.uint256


TransferEvent = Event[TransferEventType](name="Transfer")
ApprovalEvent = Event[ApprovalEventType](name="Approval")


subscriber = EventSubscriber[TransferEventType | ApprovalEventType](
    step_size=1000,
    events=[TransferEvent, ApprovalEvent],
)


@pytest.mark.asyncio(scope="session")
@pytest.mark.contract
async def test_event_subscriber_types() -> None:
    set_alchemy_key(os.environ["ALCHEMY_KEY"])
    async for event in subscriber(
        start_block=20_600_000,
        end_block=20_601_000,
        addresses=[HexAddress(HexStr("0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"))],
    ):
        assert_type(event, EventData[TransferEventType | ApprovalEventType])
        match event.event:
            case TransferEventType():
                # print("Transfer", event.event)
                pass
            case ApprovalEventType():
                # print("Approval", event.event)
                pass



from eth_rpc.types import BlockExplorer, Network, Rpcs, RpcUrl
from pydantic.networks import Url

Ethereum = Network(
    chain_id=1,
    name="Ethereum",
    native_currency="ETH",
    rpc=Rpcs(
        default=RpcUrl(
            http=Url("https://cloudflare-eth.com"),
            wss=Url("wss://mainnet.gateway.tenderly.co"),
        )
    ),
    block_explorer=BlockExplorer(
        name="Etherscan",
        url="https://etherscan.io",
        api_url="https://api.etherscan.io/api",
    ),
    alchemy_str="eth-mainnet",
    apprx_block_time=12.05,
)



from eth_rpc.types import BlockExplorer, Network, Rpcs, RpcUrl
from pydantic.networks import Url

Arbitrum = Network(
    chain_id=42161,
    name="Arbitrum",
    native_currency="ETH",
    rpc=Rpcs(
        default=RpcUrl(
            http=Url("https://arb1.arbitrum.io/rpc"),
            wss=None,
            # wss=Url("wss://test.com"),
        )
    ),
    block_explorer=BlockExplorer(
        name="Arbiscan",
        url="https://arbiscan.io",
        api_url="https://api.arbiscan.io/api",
    ),
    alchemy_str="arb-mainnet",
    apprx_block_time=0.26,
)
